import json

import pytest
from ape.api.accounts import AccountContainerAPI
from click.testing import CliRunner
from eth_typing import ChecksumAddress, HexAddress, HexStr

from ape_ledger.client import (
    LedgerEthereumAccountClient,
    LedgerEthereumAppClient,
    LedgerUsbDeviceClient,
)
from ape_ledger.hdpath import HDAccountPath, HDBasePath

TEST_ADDRESSES = [
    HexAddress(HexStr("0x0A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x1A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x2A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x3A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x4A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x5A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x6A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x7A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x8A78AAAAA2122100000b9046f0A085AB2E111113")),
    HexAddress(HexStr("0x9A78AAAAA2122100000b9046f0A085AB2E111113")),
]
TEST_ADDRESS = TEST_ADDRESSES[0]
TEST_ALIAS = "TestAlias"
TEST_HD_PATH = "m/44'/60'/0'/0/0"


@pytest.fixture
def mock_apdu(mocker):
    return mocker.MagicMock()


@pytest.fixture
def mock_device(mocker):
    return mocker.MagicMock(spec=LedgerUsbDeviceClient)


def create_test_account(mock_device, address=TEST_ADDRESS):
    address = ChecksumAddress(address)
    path = HDAccountPath(TEST_HD_PATH)
    return LedgerEthereumAccountClient(mock_device, address, path)


@pytest.fixture
def ledger_account(mock_device):
    return create_test_account(mock_device)


@pytest.fixture
def mock_ethereum_app(mocker, mock_device):
    mock = mocker.MagicMock(spec=LedgerEthereumAppClient)
    mock.hd_root_path = HDBasePath()

    def _get_address(account_id: int):
        if len(TEST_ADDRESSES) > account_id >= 0:
            address = TEST_ADDRESSES[account_id]
            return create_test_account(mock_device, address=address)

    mock.load_account.side_effect = _get_address
    return mock


@pytest.fixture
def mock_container(mocker):
    return mocker.MagicMock(spec=AccountContainerAPI)


@pytest.fixture
def runner():
    return CliRunner()


def assert_account(account_path, expected_address=TEST_ADDRESS, expected_hdpath="m/44'/60'/0'/0/0"):
    with open(account_path) as account_file:
        account_data = json.load(account_file)
        assert account_data["address"] == expected_address
        assert account_data["hdpath"] == expected_hdpath



from ape.exceptions import AccountsError


class LedgerAccountException(AccountsError):
    """
    An error that occurs in the ape Ledger plugin.
    """


class LedgerSigningError(LedgerAccountException):
    """
    An error that occurs when signing a message or transaction
    using the Ledger plugin.
    """


class LedgerUsbError(LedgerAccountException):
    def __init__(self, message: str, status: int = 0):
        self.status = status
        super().__init__(message)


class LedgerTimeoutError(LedgerUsbError):
    """
    Raised when the Ledger client times-out waiting for a response from the device.
    """

    def __init__(self, timeout: int):
        message = (
            f"Timeout waiting device response (timeout={timeout}).\n"
            "Make sure the Ledger device is not busy with another task."
        )
        super().__init__(message)



import json
from pathlib import Path
from typing import Iterator, Optional, Union

import click
import rlp  # type: ignore
from ape.api import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, TransactionSignature
from ape_ethereum.transactions import TransactionType
from eth_account.messages import SignableMessage
from hexbytes import HexBytes

from ape_ledger.client import LedgerEthereumAccountClient, connect_to_ethereum_account
from ape_ledger.exceptions import LedgerSigningError
from ape_ledger.hdpath import HDAccountPath
from ape_ledger.objects import DynamicFeeTransaction, StaticFeeTransaction


class AccountContainer(AccountContainerAPI):
    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for account_file in self._account_files:
            yield LedgerAccount(container=self, account_file_path=account_file)

    def __setitem__(self, address: AddressType, account: AccountAPI):
        raise NotImplementedError()

    def __delitem__(self, address: AddressType):
        raise NotImplementedError()

    @property
    def _account_files(self) -> Iterator[Path]:
        return self.data_folder.glob("*.json")

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._account_files:
            yield p.stem

    def __len__(self) -> int:
        return len([*self._account_files])

    def save_account(self, alias: str, address: str, hd_path: str):
        """
        Save a new Ledger account to your ape configuration.
        """
        account_data = {"address": address, "hdpath": hd_path}
        path = self.data_folder.joinpath(f"{alias}.json")
        path.write_text(json.dumps(account_data))

    def delete_account(self, alias: str):
        path = self.data_folder.joinpath(f"{alias}.json")

        if path.exists():
            path.unlink()


def _echo_object_to_sign(obj: Union[TransactionAPI, SignableMessage]):
    click.echo(f"{obj}\nPlease follow the prompts on your device.")


class LedgerAccount(AccountAPI):
    account_file_path: Path

    # Optional because it's lazily loaded
    account_client: Optional[LedgerEthereumAccountClient] = None

    @property
    def alias(self) -> str:
        return self.account_file_path.stem

    @property
    def address(self) -> AddressType:
        ecosystem = self.network_manager.get_ecosystem("ethereum")
        return ecosystem.decode_address(self.account_file["address"])

    @property
    def hdpath(self) -> HDAccountPath:
        raw_path = self.account_file["hdpath"]
        return HDAccountPath(raw_path)

    @property
    def account_file(self) -> dict:
        return json.loads(self.account_file_path.read_text())

    @property
    def _client(self) -> LedgerEthereumAccountClient:
        if self.account_client is None:
            self.account_client = connect_to_ethereum_account(self.address, self.hdpath)
        return self.account_client

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        version = msg.version
        if version == b"E":
            _echo_object_to_sign(msg)
            signed_msg = self._client.sign_personal_message(msg.body)
        elif version == b"\x01":
            _echo_object_to_sign(msg)
            signed_msg = self._client.sign_typed_data(msg.header, msg.body)
        else:
            raise LedgerSigningError(
                f"Unsupported message-signing specification, (version={version!r})."
            )

        v, r, s = signed_msg
        return MessageSignature(v, r, s)

    def sign_transaction(self, txn: TransactionAPI, **kwargs) -> Optional[TransactionAPI]:
        txn_type = TransactionType(txn.type)  # In case it is not enum
        if txn_type == TransactionType.STATIC:
            serializable_txn = StaticFeeTransaction(**txn.dict())
            txn_bytes = rlp.encode(serializable_txn, StaticFeeTransaction)
        else:
            serializable_txn = DynamicFeeTransaction(**txn.dict())
            version_byte = bytes(HexBytes(TransactionType.DYNAMIC.value))
            txn_bytes = version_byte + rlp.encode(serializable_txn, DynamicFeeTransaction)

        _echo_object_to_sign(txn)
        v, r, s = self._client.sign_transaction(txn_bytes)

        chain_id = txn.chain_id
        # NOTE: EIP-1559 transactions don't pack 'chain_id' with 'v'.
        if txn_type != TransactionType.DYNAMIC and (chain_id * 2 + 35) + 1 > 255:
            ecc_parity = v - ((chain_id * 2 + 35) % 256)
            v = (chain_id * 2 + 35) + ecc_parity

        txn.signature = TransactionSignature(v, r, s)
        return txn



import pytest
from ape import accounts
from ape._cli import cli

from ape_ledger.hdpath import HDBasePath

from .conftest import TEST_ADDRESS, TEST_HD_PATH, assert_account


def _get_container():
    return accounts.containers["ledger"]


TEST_ALIAS = "__integration_test_alias__"
TEST_ACCOUNT_PATH = _get_container().data_folder.joinpath(f"{TEST_ALIAS}.json")


@pytest.fixture
def mock_device_connection(mocker, mock_ethereum_app):
    patch = mocker.patch("ape_ledger._cli.connect_to_ethereum_app")
    patch.return_value = mock_ethereum_app
    return patch


@pytest.fixture
def existing_account(runner):
    try:
        container = _get_container()
        container.save_account(TEST_ALIAS, TEST_ADDRESS, TEST_HD_PATH)
        yield
    finally:
        _clean_up(runner)


@pytest.fixture(autouse=True)
def clean_after(runner):
    try:
        yield
    finally:
        _clean_up(runner)


def _clean_up(runner):
    runner.invoke(cli, ["ledger", "delete", TEST_ALIAS], input="y")


def _get_account_path(alias=TEST_ALIAS):
    container = _get_container()
    return container.data_folder.joinpath(f"{alias}.json")


@pytest.mark.parametrize("cmd", (["ledger", "list"], ["accounts", "list", "--all"]))
def test_list(runner, existing_account, cmd):
    result = runner.invoke(cli, cmd)
    assert result.exit_code == 0, result.output
    assert TEST_ALIAS in result.output
    assert TEST_ADDRESS.lower() in result.output.lower()


def test_add(runner, mock_device_connection):
    selected_account_id = 0
    result = runner.invoke(cli, ["ledger", "add", TEST_ALIAS], input=str(selected_account_id))
    assert result.exit_code == 0, result.output
    assert (
        f"SUCCESS: Account '{TEST_ADDRESS}' successfully added with alias '{TEST_ALIAS}'."
        in result.output
    )

    container = _get_container()
    expected_path = container.data_folder.joinpath(f"{TEST_ALIAS}.json")
    expected_hd_path = f"m/44'/60'/{selected_account_id}'/0/0"
    assert_account(expected_path, expected_hdpath=expected_hd_path)


def test_add_when_hd_path_specified(runner, mock_ethereum_app, mock_device_connection):
    test_hd_path = "m/44'/60'/0'"
    mock_ethereum_app.hd_root_path = HDBasePath(test_hd_path)

    selected_account_id = 0
    result = runner.invoke(
        cli,
        ["ledger", "add", TEST_ALIAS, "--hd-path", test_hd_path],
        input=str(selected_account_id),
    )
    assert result.exit_code == 0, result.output
    assert (
        f"SUCCESS: Account '{TEST_ADDRESS}' successfully added with alias '{TEST_ALIAS}'."
        in result.output
    )

    expected_path = TEST_ACCOUNT_PATH
    expected_hd_path = f"m/44'/60'/0'/{selected_account_id}"
    assert_account(expected_path, expected_hdpath=expected_hd_path)


def test_add_alias_already_exists(
    runner, mock_ethereum_app, mock_device_connection, existing_account
):
    result = runner.invoke(cli, ["ledger", "add", TEST_ALIAS], input="0")
    assert result.exit_code == 1, result.output
    assert (
        f"ERROR: (AliasAlreadyInUseError) Account with alias '{TEST_ALIAS}' already in use."
        in result.output
    )


def test_delete(runner, existing_account):
    result = runner.invoke(cli, ["ledger", "delete", TEST_ALIAS])
    assert result.exit_code == 0, result.output
    assert f"SUCCESS: Account '{TEST_ALIAS}' has been removed" in result.output


def test_delete_account_not_exists(runner):
    result = runner.invoke(cli, ["ledger", "delete", TEST_ALIAS])
    assert result.exit_code == 2
    assert f"'{TEST_ALIAS}' is not one of" in result.output



import ape_ledger.client as client_module


def test_device_connects_once(mock_device):
    client_module.device_manager._device = mock_device
    assert client_module.device_manager.device == mock_device



from typing import Optional, Union

from eth_typing import HexStr
from eth_utils import add_0x_prefix, decode_hex
from rlp import Serializable  # type: ignore
from rlp.sedes import BigEndianInt, Binary, CountableList  # type: ignore
from rlp.sedes import List as ListSedes
from rlp.sedes import big_endian_int, binary  # type: ignore

"""Inspired from https://github.com/mikeshultz/ledger-eth-lib"""


def _encode_hex(receiver: Optional[Union[str, bytes]] = None) -> bytes:
    if receiver is None:
        return b""

    if isinstance(receiver, str):
        add_0x_prefix(HexStr(receiver))
        # Note: We are encoding to bytes, nevermind the name of this method 'decode_hex'.
        return decode_hex(receiver)

    return receiver


# Define typed transaction common sedes.
# [[{20 bytes}, [{32 bytes}...]]...], where ... means “zero or more of the thing to the left”.
access_list_sede_type = CountableList(
    ListSedes(
        [
            Binary.fixed_length(20, allow_empty=False),
            CountableList(BigEndianInt(32)),
        ]
    ),
)


class StaticFeeTransaction(Serializable):
    fields = [
        ("nonce", big_endian_int),
        ("gas_sprice", big_endian_int),
        ("gas_limit", big_endian_int),
        ("receiver", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("chain_id", big_endian_int),
        # Expected nine elements as part of EIP-155 transactions
        ("ignored_1", big_endian_int),
        ("ignored_2", big_endian_int),
    ]

    def __init__(
        self,
        nonce: int,
        gasPrice: int,
        gas: int,
        value: int,
        data: bytes,
        chainId: int,
        to: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            nonce,
            gasPrice,
            gas,
            _encode_hex(to),
            value,
            data,
            chainId,
            0,
            0,
        )


class DynamicFeeTransaction(Serializable):
    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("max_priority_fee", big_endian_int),
        ("max_fee", big_endian_int),
        ("gas_limit", big_endian_int),
        ("receiver", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("access_list", access_list_sede_type),
    ]

    def __init__(
        self,
        nonce: int,
        maxFeePerGas: int,
        maxPriorityFeePerGas: int,
        gas: int,
        value: int,
        data: bytes,
        chainId: int,
        to: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            chainId,
            nonce,
            maxPriorityFeePerGas,
            maxFeePerGas,
            gas,
            _encode_hex(to),
            value,
            data,
            [],
        )



"""
Implementation inspired from
https://github.com/vegaswap/ledgertools/blob/master/ledger_usb.py
"""
import struct
import time
from typing import List, Optional, Tuple

import hid  # type: ignore
from eth_typing.evm import ChecksumAddress
from eth_utils import to_checksum_address

from ape_ledger.exceptions import LedgerTimeoutError, LedgerUsbError
from ape_ledger.hdpath import HDAccountPath, HDBasePath

_CHANNEL_ID = 0x0101


class _CmdTag:
    APDU = 0x05
    PING = 0x02


class _PacketData:
    # Packet HEADER is defined at
    # 1. Communication channel ID (big endian) value is always CHANNEL_ID
    # 2. Command tag (TAG_APDU or TAG_PING)
    # 3. Packet sequence index (big endian) starting at 0x00

    HEADER = struct.pack(">HBH", _CHANNEL_ID, _CmdTag.APDU, 0x00)
    SIZE = 64  # in bytes
    FREE = SIZE - len(HEADER)


class APDUStatus:
    """Device-returned status words.

    Source:
        https://github.com/LedgerHQ/app-ethereum/blob/master/doc/ethapp.asc#status-words
    """

    TX_TYPE_UNSUPPORTED = 0x6501
    OUTPUT_BUFFER_TOO_SMALL = 0x6502
    PLUGIN_ERROR = 0x6503
    INT_CONVERSION_ERROR = 0x6504
    INCORRECT_LENGTH = 0x6700
    CANCELED_BY_USER = 0x6982
    DECLINED = 0x6985
    INVALID_DATA = 0x6A80
    INCORRECT_PARAMETER = 0x6A80
    APP_SLEEP = 0x6804
    APP_NOT_STARTED = 0x6D00
    DEVICE_LOCKED = 0x6B0C
    OK = 0x9000

    @classmethod
    def to_status_message(cls, status: int) -> str:
        """
        Convert a byte word to the English equivalent.
        """

        internal_error = "Internal error"
        if 0x6F00 <= status <= 0x6FFF:
            return internal_error

        mapping = {
            cls.TX_TYPE_UNSUPPORTED: "TransactionType not supported",
            cls.OUTPUT_BUFFER_TOO_SMALL: "Output buffer too small for chainId conversion",
            cls.PLUGIN_ERROR: "Plugin error",
            cls.INT_CONVERSION_ERROR: "Failed to convert from int256",
            cls.INCORRECT_LENGTH: "Incorrect length",
            cls.CANCELED_BY_USER: "Security status not satisfied. Canceled by user.",
            cls.DECLINED: "User declined on device",
            cls.INVALID_DATA: "Invalid data",
            cls.INCORRECT_PARAMETER: "Incorrect parameter P1 or P2",
            cls.APP_SLEEP: "Ethereum app not ready on device",
            cls.APP_NOT_STARTED: "Ethereum app not started on device",
            cls.DEVICE_LOCKED: "The device is locked",
            cls.OK: "OK",
        }

        return mapping.get(status, internal_error)


class _Code:
    """Device command codes."""

    CLA = 0xE0
    INS_GET_PUBLIC_KEY = 0x02
    INS_SIGN_TX = 0x04
    INS_GET_APP_CONFIGURATION = 0x06
    INS_SIGN_PERSONAL_MESSAGE = 0x08
    INS_SIGN_EIP_712 = 0x0C
    P1_CONFIRM = 0x01
    P1_NON_CONFIRM = 0x00
    P2_NO_CHAINCODE = 0x00
    P2_CHAINCODE = 0x01
    P1_FIRST = 0x00
    P1_MORE = 0x80


_LEDGER_VENDOR_ID = 0x2C97
_LEDGER_USAGE_PAGE_ID = 0xFFA0


class APDUBuilder:
    def __init__(self, ins: int, p1: int = _Code.P1_FIRST, p2: int = _Code.P2_NO_CHAINCODE):
        self.apdu = struct.pack(">BBBB", _Code.CLA, ins, p1, p2)

    def append(self, path: bytes, payload: bytes):
        length = len(payload) + len(path)
        self.apdu += struct.pack(">B", length)
        self.apdu += path + payload


def _wrap_apdu(command: bytes) -> List[bytes]:
    """Return a list of packet to be sent to the device"""

    packets = []
    header = struct.pack(">H", len(command))
    command = header + command
    chunks = [command[i : i + _PacketData.FREE] for i in range(0, len(command), _PacketData.FREE)]

    # Create a packet for each command chunk
    for packet_id in range(len(chunks)):
        header = struct.pack(">HBH", _CHANNEL_ID, _CmdTag.APDU, packet_id)
        packet = header + chunks[packet_id]
        packet.ljust(_PacketData.SIZE, bytes([0x0]))
        packets.append(packet)

    return packets


def _unwrap_apdu(
    packet: bytes,
) -> Tuple[Optional[int], Optional[bytes], Optional[int], Optional[int], Optional[bytes]]:
    """
    Given a packet from the device, extract and return relevant info
    """
    if not packet:
        return None, None, None, None, None

    (channel, tag, packet_id, reply_size) = struct.unpack(">HBHH", packet[:7])

    if packet_id == 0:
        return channel, tag, packet_id, reply_size, packet[7:]

    return channel, tag, packet_id, None, packet[5:]


def _get_hid_device_path() -> Optional[str]:
    hid_info = {}
    for hid_info in hid.enumerate(0, 0):
        if hid_info.get("vendor_id") != _LEDGER_VENDOR_ID:
            continue

        interface_number = hid_info.get("interface_number")
        usage_page = hid_info.get("usage_page")
        if interface_number == 0 or usage_page == _LEDGER_USAGE_PAGE_ID:
            # Found
            break

    return hid_info.get("path")


def _get_device(hid_path: str) -> hid.device:
    device = hid.device()

    try:
        device.open_path(hid_path)
    except OSError as err:
        message = (
            "Unable to open HID path. "
            "Make sure you have your device unlocked via the passcode "
            "and have the Ethereum app open."
        )
        raise LedgerUsbError(message) from err

    device.set_nonblocking(True)
    return device


def _handle_reply_status(reply: bytes):
    (status,) = struct.unpack(">H", reply[-2:])

    if status == APDUStatus.OK:
        return reply[:-2]

    message = APDUStatus.to_status_message(status)
    raise LedgerUsbError(message, status=status)


def _verify_channel_and_tag(channel: Optional[int], tag: Optional[bytes]):
    if channel != _CHANNEL_ID:
        raise LedgerUsbError(f"Invalid channel '{channel}'. Expecting '{_CHANNEL_ID}'.")
    elif tag != _CmdTag.APDU:
        raise LedgerUsbError(f"Invalid tag '{tag!r}'. Expecting '{_CmdTag.APDU!r}'.")


class LedgerUsbDeviceClient:
    """
    This class is a APDU client for the Ledger device.
    It abstracts away communication with the device via it's
    :meth:`~ape_ledger.ledger.LedgerUsbDevice.exchange()` method.

    References:
    - https://github.com/ethereum/go-ethereum/blob/master/accounts/usbwallet/ledger.go
    """

    _exchange_timeout = 60

    def __init__(self, hid_device):
        self._device = hid_device

    def exchange(self, apdu: bytes) -> bytes:
        packets = _wrap_apdu(apdu)
        for packet in packets:
            self._device.write(packet)

        reply = self._receive_reply()
        return _handle_reply_status(reply)

    def _receive_reply(self) -> bytes:
        """
        Receive reply, size of reply is contained in first packet.
        """
        reply: bytes = b""
        reply_min_size = 2
        reply_start = time.time()
        while True:
            packet = bytes(self._device.read(64))
            (channel, tag, index, size, data) = _unwrap_apdu(packet)

            channel = self._wait_for_channel(channel, reply_start)
            if not channel:
                continue

            _verify_channel_and_tag(channel, tag)

            # Size is not None only on first reply
            reply_min_size = size or reply_min_size
            if data:
                reply += data

            # Check if we have received all the reply from device
            if len(reply) > reply_min_size:
                reply = bytes(reply[:reply_min_size])
                break

        return reply

    def _wait_for_channel(self, channel, reply_start):
        if channel:
            return channel

        if reply_start + self._exchange_timeout < time.time():
            raise LedgerTimeoutError(self._exchange_timeout)

        time.sleep(0.01)


def _to_vrs(reply: bytes) -> Tuple[int, bytes, bytes]:
    """
    Breaks a byte message into 3 chunks vrs,
    where `v` is 1 byte, `r` is 32 bytes, and `s` is 32 bytes.
    """
    if not reply:
        raise LedgerUsbError("No data in reply.")

    v = reply[0]  # 1 byte
    r = reply[1:33]  # 32 bytes
    s = reply[33:65]  # 32 bytes
    return v, r, s


class LedgerEthereumAccountClient:
    """
    This class represents an account on the Ledger device when you know the full
    account HD path.
    """

    def __init__(
        self,
        client: LedgerUsbDeviceClient,
        address: ChecksumAddress,
        account_hd_path: HDAccountPath,
    ):
        self._client = client
        self._address = address
        self._account_hd_path = account_hd_path
        self._path_bytes = b""  # cached calculation

    def __str__(self):
        return self._address

    @property
    def address(self) -> str:
        return self._address

    @property
    def path_bytes(self) -> bytes:
        if self._path_bytes == b"":
            self._path_bytes = self._account_hd_path.as_bytes()

        return self._path_bytes

    def sign_personal_message(self, message_bytes: bytes) -> Tuple[int, bytes, bytes]:
        """
        Sign an Ethereum message only following the EIP 191 specification and using
        your Ledger device. You will need to follow the prompts on the device
        to validate the message data.

        APDU breakdown:
          Number of BIP 32 derivations to perform (max 10) - 1 byte
          First derivation index (big endian)              - 4 bytes
          ...                                              - 4 bytes
          Last derivation index (big endian)               - 4 bytes
          Message length                                   - 4 bytes
          Message chunk                                    - variable
        """

        encoded_message = struct.pack(">I", len(message_bytes))
        encoded_message += message_bytes
        builder = APDUBuilder(_Code.INS_SIGN_PERSONAL_MESSAGE)
        builder.append(self.path_bytes, encoded_message)
        reply = self._client.exchange(builder.apdu)
        return _to_vrs(reply)

    def sign_typed_data(self, domain_hash: bytes, message_hash: bytes) -> Tuple[int, bytes, bytes]:
        """
        Sign an Ethereum message following the EIP 712 specification.

        APDU breakdown:
          Number of BIP 32 derivations to perform (max 10) - 1 byte
          First derivation index (big endian)              - 4 bytes
          ...                                              - 4 bytes
          Last derivation index (big endian)               - 4 bytes
          domain hash                                      - 32 bytes
          message hash                                     - 32 bytes
        """

        encoded_message = domain_hash + message_hash
        builder = APDUBuilder(_Code.INS_SIGN_EIP_712)
        builder.append(self.path_bytes, encoded_message)
        reply = self._client.exchange(builder.apdu)
        return _to_vrs(reply)

    def sign_transaction(self, txn: bytes) -> Tuple[int, bytes, bytes]:
        """
        Sign a transaction using your Ledger device. You will need to follow
        the prompts on the device to validate the transaction data.

        APDU breakdown:
          Number of BIP 32 derivations to perform (max 10) - 1 byte
          First derivation index (big endian)              - 4 bytes
          ...                                              - 4 bytes
          Last derivation index (big endian)               - 4 bytes
          RLP transaction chunk                            - arbitrary
        """

        payload = self.path_bytes + txn
        chunks = [payload[i : i + 255] for i in range(0, len(payload), 255)]  # noqa: E203
        apdu_param1 = _Code.P1_FIRST
        reply = None

        for chunk in chunks:
            builder = APDUBuilder(_Code.INS_SIGN_TX, p1=apdu_param1)
            builder.append(b"", chunk)
            reply = self._client.exchange(builder.apdu)
            apdu_param1 = _Code.P1_MORE

        if not reply:
            raise LedgerUsbError("Signing transaction failed - received 0 bytes in reply.")

        return _to_vrs(reply)


class LedgerEthereumAppClient:
    """
    This class is able to get accounts from your Ethereum wallet.
    """

    def __init__(self, client: LedgerUsbDeviceClient, hd_path: HDBasePath):
        self._client = client
        self.hd_root_path = hd_path

    def load_account(self, account_id: int) -> LedgerEthereumAccountClient:
        path_bytes = self.hd_root_path.get_account_path(account_id).as_bytes()
        builder = APDUBuilder(_Code.INS_GET_PUBLIC_KEY, p1=_Code.P1_NON_CONFIRM)
        builder.append(path_bytes, b"")
        account_data = self._client.exchange(builder.apdu)
        offset = 1 + account_data[0]
        address = account_data[offset + 1 : offset + 1 + account_data[offset]]
        address_checksum = to_checksum_address(address.decode())
        return LedgerEthereumAccountClient(self._client, address_checksum, path_bytes)


def connect_to_ethereum_account(
    address: ChecksumAddress, hd_account_path: HDAccountPath
) -> LedgerEthereumAccountClient:
    """
    Create an account client using an active device connection.
    """

    return LedgerEthereumAccountClient(device_manager.device, address, hd_account_path)


def connect_to_ethereum_app(hd_path: HDBasePath) -> LedgerEthereumAppClient:
    """
    Create a client that is able to create account clients,
    using an active device connection.
    """

    return LedgerEthereumAppClient(device_manager.device, hd_path)


class DeviceManager:
    _device: Optional[LedgerUsbDeviceClient] = None
    _client_cls = LedgerUsbDeviceClient

    @property
    def device(self) -> LedgerUsbDeviceClient:
        """
        Create a Ledger device client and connect to it.
        """

        if self._device:
            return self._device

        hid_path = _get_hid_device_path()
        if hid_path is None:
            raise LedgerUsbError("No Ledger USB device found.")

        device = _get_device(hid_path)
        self._device = self._client_cls(device)
        return self._device


device_manager = DeviceManager()


__all__ = [
    "connect_to_ethereum_account",
    "connect_to_ethereum_app",
    "device_manager",
    "LedgerEthereumAccountClient",
    "LedgerEthereumAppClient",
    "LedgerUsbDeviceClient",
]
